static
void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
{
   _ntl_ulong hs0, hs1;
   _ntl_ulong hl2[2];

   hs0 = a[0] ^ a[1];
   hs1 = b[0] ^ b[1];

   mul1(c, a[0], b[0]);
   mul1(c+2, a[1], b[1]);
   mul1(hl2, hs0, hs1);


   hl2[0] = hl2[0] ^ c[0] ^ c[2];
   hl2[1] = hl2[1] ^ c[1] ^ c[3];

   c[1] ^= hl2[0];
   c[2] ^= hl2[1];
}

/* an = bn = n > 0 */
static inline void
mul_basecase_n (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                long n)
{
  long i;

  switch (n)
    {
    case 1:
      mul1 (c, a[0], b[0]);
      return;
    case 2:
      mul2 (c, a, b); // faster than default
      return;
    default:
      c[n] = Mul1 (c, a, n, b[0]);
      for (i = 1; i < n; i++)
        c[n + i] = AddMul1 (c + i, c + i, a, n, b[i]);
      return;
    }
}

static
void KarMul(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, 
            long len, _ntl_ulong *stk)
{
   if (len < MUL_KARA_THRESHOLD)
     {
       mul_basecase_n (c, a, b, len);
       return;
     }

   long ll, lh, i, ll2, lh2;
   const _ntl_ulong *a0, *a1, *b0, *b1;
   _ntl_ulong *a01, *b01, *h;

   lh = len >> 1;
   ll = (len+1) >> 1;

   ll2 = ll << 1;
   lh2 = lh << 1;

   a01 = stk;  stk += ll+1;
   b01 = stk;  stk += ll+1;
   h = stk; stk += ll2+1;

   a0 = a;
   a1 = a+ll;
   b0 = b;
   b1 = b+ll;

   KarMul(c, a0, b0, ll, stk);
   KarMul(c+ll2, a1, b1, lh, stk);

   for (i = 0; i < lh; i++) {
      a01[i] = a[i] ^ a[i+ll];
      b01[i] = b[i] ^ b[i+ll];
   }

   if (lh < ll) {
      a01[lh] = a[lh];
      b01[lh] = b[lh];
   }

   KarMul(h, a01, b01, ll, stk);

   for (i = 0; i < ll2; i++)
      h[i] ^= c[i];

   for (i = 0; i < lh2; i++)
      h[i] ^= c[i+ll2];

   for (i = 0; i < ll2; i++)
      c[i+ll] ^= h[i];
}

/* c <- a + b */
static
void Add (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, long n)
{
  long i;
  for (i = 0; i < n; i++)
    c[i] = a[i] ^ b[i];
}

/* c <- c + a + b */
static
void Add3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b, long n)
{
  long i;
  for (i = 0; i < n; i++)
    c[i] ^= a[i] ^ b[i];
}

/* c <- a + x * b, return carry out */
static
_ntl_ulong AddLsh1 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 1) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 1);
      c[i] = t;
    }
  return cy;
}

/* c <- a + x^4 * b, return carry out */
static
_ntl_ulong AddLsh4 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 4) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 4);
      c[i] = t;
    }
  return cy;
}

/* c <- x * c, return carry out */
static
_ntl_ulong Lsh1 (_ntl_ulong *c, long n, _ntl_ulong cy)
{
  _ntl_ulong t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = (c[i] << 1) | cy;
      cy = c[i] >> (NTL_BITS_PER_LONG - 1);
      c[i] = t;
    }
  return cy;
}

/* c <- c/x, return carry */
static
_ntl_ulong Rsh1 (_ntl_ulong *c, long n)
{
  _ntl_ulong cy = 0, t;
  long i;
  for (i = n - 1; i >= 0; i--)
    {
      t = (c[i] >> 1) | (cy << (NTL_BITS_PER_LONG - 1));
      cy = c[i] & 1UL;
      c[i] = t;
    }
#ifdef DEBUG
  if (cy) abort();
#endif
  return cy;
}

/* c <- a + cy, return carry out (0 for n > 0, cy for n=0) */
static
_ntl_ulong Add1 (_ntl_ulong *c, const _ntl_ulong *a, long n, _ntl_ulong cy)
{
  if (n)
    {
      long i;
      c[0] = a[0] ^ cy;
      for (i = 1; i < n; i++)
        c[i] = a[i];
      return 0;
    }
  else
    return cy;
}

/* let c = q*(1+x^2) + X^n*r with X = x^NTL_BITS_PER_LONG and deg(r) < 2
   then c <- q, returns r.
   (Algorithm from Michel Quercia.)
*/
static _ntl_ulong DivOnePlusX2 (_ntl_ulong *c, long n)
{
  _ntl_ulong t = 0;
  long i;

  for (i = 0; i < n; i++) {
    t ^= c[i];
    t ^= t << 2;
    t ^= t << 4;
    t ^= t << 8;
    t ^= t << 16;
#if (NTL_BITS_PER_LONG == 64)
    t ^= t << 32;
#elif (NTL_BITS_PER_LONG != 32)
#error "NTL_BITS_PER_LONG should be 32 or 64"
#endif
    c[i] = t;
    t >>= (NTL_BITS_PER_LONG - 2);
  }
  return t;
}

/*
  c := proc(t) c0+t*c1+t^2*c2+t^3*c3+t^4*c4 end:
  solve ({c(1) = ab1, c(x) = abx, expand(x^4*c(1/x)) = abix}, 
          {c1, c2, c3});
  After ab1 <- ab1 - c0 - c4,
        abx <- (abx - c0 - c4*x^4)/x
        abix <- (abix - c0*x^4 - c4)/x
  we have:
     B := c1*x^2 + c2*x + c3:
     C := c1 + c2 + c3:
     D := c1 + c2*x + c3*x^2:
  solve({C=c1+c2+c3, D=c1+c2*x+c3*x^2, B=c1*x^2+c2*x+c3}, {c1, c2, c3});
  map(proc(x) lhs(x)=evala(Factor(numer(rhs(x)) mod 2))/
       evala(Factor(denom(rhs(x)) mod 2)) end, %);
  t2 := (B + D) / (1+x)^2:          # c1+c3
  c2 := C + t2                      # c2
  c1 := (x*c2 + B + t2)/(1+x)^2:    # c1
  c3 = c1 + t2                      # c3
 */
static
void ToomMul (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
              long n, _ntl_ulong *stk)
{
  while (n && a[n-1] == 0 && b[n-1] == 0)
    {
      c[2*n-1] = 0;
      c[2*n-2] = 0;
      n --;
    }

  if (n < MUL_TOOM_THRESHOLD)
    return KarMul (c, a, b, n, stk);

  long k = (n + 2) / 3; /* ceil(n/3) */
  long r = n - 2 * k;
  _ntl_ulong cy;
  _ntl_ulong *A = c;
  _ntl_ulong *B = stk + 2 * k + 1;
  _ntl_ulong *C = c + 2 * k;
  _ntl_ulong *D = stk;
  _ntl_ulong *E = c + 4 * k;

  cy = AddLsh1 (c, a + k, a + 2 * k, r);   /* a1 + x a2 */
  cy = Add1 (c + r, a + k + r, k - r, cy);
  c[k] = (cy << 1) ^ AddLsh1 (c, a, c, k); /* a0 + x a1 + x^2 a2 */
  cy = AddLsh1 (c + k + 1, b + k, b + 2 * k, r);
  cy = Add1 (c + k + r + 1, b + k + r, k - r, cy);
  c[2 * k + 1] = (cy << 1) ^ AddLsh1 (c + k + 1, b, c + k + 1, k);
  ToomMul (D, c, c + k + 1, k + 1, stk + 2 * k + 2); /* D : 2*k+1 */

  c[k] = AddLsh1 (c, a + k, a, k);   /* a1 + x a0 */
  cy = AddLsh1 (c, a + 2 * k, c, r); /* x^2 a0 + x a1 + a2 */
  c[k] = (c[k] << 1) ^ Lsh1 (c + r, k - r, cy);
  c[2 * k + 1] = AddLsh1 (c + k + 1, b + k, b, k); /* b1 + x b0 */
  cy = AddLsh1 (c + k + 1, b + 2 * k, c + k + 1, r);
  c[2 * k + 1] = (c[2 * k + 1] << 1) ^ Lsh1 (c + k + r + 1, k - r, cy);
  ToomMul (B, c, c + k + 1, k + 1, stk + 4 * k + 3); /* B : 2*k+1 */

  Add (c, a, a + k, k);
  Add (c, c, a + 2 * k, r);
  Add (stk + 4*k+2, b, b + k, k);
  Add (stk + 4*k+2, stk + 4*k+2, b + 2 * k, r);
  ToomMul (C, c, stk + 4*k+2, k, stk + 5*k+2); /* C : 2*k */

  ToomMul (A, a, b, k, stk+4*k+2); /* A : 2*k */

  ToomMul (E, a + 2 * k, b + 2 * k, r, stk+4*k+2); /* E : 2*r */

  /* C <- C - A - E  = c1 + c2 + c3 */
  Add3 (C, A, E, 2 * r);
  Add (C + 2 * r, C + 2 * r, A + 2 * r, 2 * (k - r));

  /* B <- (B - A*x^4 - E)/x */
  Add (B, B, E, 2 * r);
  B[2*k] ^= AddLsh4 (B, B, A, 2 * k);
  Rsh1 (B, 2*k+1); /* (a0*b1+a1*b0)*x^2+(a0*b2+a1*b1+a2*b0)*x+a1*b2+a2*b1 */

  /* D <- (D - A - E*x^4)/x */
  Add (D, D, A, 2 * k);
  D[2*r] ^= AddLsh4 (D, D, E, 2 * r);
  Rsh1 (D, 2*k+1); /* (a1*b2+a2*b1)*x^2+(a0*b2+a1*b1+a2*b0)*x+a0*b1+a1*b0 */

  /* D <- (B + D)/(1+x)^2 */
  Add (D, D, B, 2*k+1); /* (a1*b2+a2*b1+a0*b1+a1*b0)*(x^2+1) */
  DivOnePlusX2 (D, 2*k+1);
  //  if (D[2*k]) abort();

  /* C <- C + t2 */
  Add (C, C, D, 2*k);

  /* B <- (B + t2 + x*c2) / (1+x)^2 */
  Add (B, B, D, 2*k);
  B[2*k] ^= AddLsh1 (B, B, C, 2*k);
  DivOnePlusX2 (B, 2*k+1);
  //  if (B[2*k]) abort();

  Add3 (c + 3 * k, B, D, 2 * k);
  Add (c + k, c + k, B, 2 * k);
}

/********************************************************************
 * Below this line, experimental code
 * (C) 2007 Marco Bodrato <optimaltoom@bodrato.it>
 * This code is released under GPL 2.0 licence.
 *
 * Reference: http://bodrato.it/papers/#WAIFI2007
 *
 * "Towards Optimal Toom-Cook Multiplication for Univariate and
 * Multivariate Polynomials in Characteristic 2 and 0." by Marco
 * BODRATO; in C.Carlet and B.Sunar, editors, "WAIFI'07 proceedings",
 * LNCS 4547, pp. 116-133. Springer, Madrid, Spain, June 21-22, 2007.
 */

/* Thanks to Paul Zimmermann for pointing out some small bugs, and the
 * limitation below for the two algorithms. After his proofreading, this
 * implementation passes all my tests. Hurra Paul !-)
 */
#if (MUL_TOOM4_THRESHOLD < 30)
#error "MUL_TOOM4_THRESHOLD should be at least 30"
#endif
#if (MUL_TOOM_THRESHOLD < 17)
#error "MUL_TOOM_THRESHOLD should be at least 17"
#endif

/* c <- x * a + x^2 * b, return carry out */
static
_ntl_ulong AddLsh1Lsh2 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = (a[i]<<1) ^ ((b[i] << 2) | cy);
      cy =  (a[i] >> (NTL_BITS_PER_LONG - 1)) ^ (b[i] >> (NTL_BITS_PER_LONG - 2));
      c[i] = t;
    }
  return cy;
}

/* c <- ( a + b )/x, return carry */
static
_ntl_ulong Rsh1Add (_ntl_ulong *c,  const _ntl_ulong *a, const _ntl_ulong *b, long n)
{
  _ntl_ulong cy = 0, t, t2;
  long i;
  for (i = n - 1; i >= 0; i--)
    {
      t2 = a[i] ^ b[i];
      t = (t2 >> 1) | (cy << (NTL_BITS_PER_LONG - 1));
      cy = t2 & 1UL;
      c[i] = t;
    }
  return cy;
}

/* c <- a + x^3 * b, return carry out */
static
_ntl_ulong AddLsh3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 3) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 3);
      c[i] = t;
    }
  return cy;
}

/* let c = q*(1+x) + X^n*r with X = x^NTL_BITS_PER_LONG and deg(r) < 1
   then c <- q, returns r.
   (Algorithm from Michel Quercia.)
*/
static _ntl_ulong DivOnePlusX (_ntl_ulong *c, long n)
{
  _ntl_ulong t = 0;
  long i;

  for (i = 0; i < n; i++) {
    t ^= c[i];
    t ^= t << 1;
    t ^= t << 2;
    t ^= t << 4;
    t ^= t << 8;
    t ^= t << 16;
#if (NTL_BITS_PER_LONG == 64)
    t ^= t << 32;
#elif (NTL_BITS_PER_LONG != 32)
#error "NTL_BITS_PER_LONG should be 32 or 64"
#endif
    c[i] = t;
    t >>= (NTL_BITS_PER_LONG - 1);
  }
  return t;
}

/*
\\ gp-pari check code.
default(echo, 1);

A = (a2*x^2 + a1*x + a0)*Mod(1,2)
B = (b2*x^2 + b1*x + b0)*Mod(1,2)
C = A * B
c0 = polcoeff(C, 0)
c1 = polcoeff(C, 1)
c2 = polcoeff(C, 2)
c3 = polcoeff(C, 3)
c4 = polcoeff(C, 4)

\\ --- Evaluation phase. 10 add, 4 shift, 5 mul.

W0 = (a2*y^2+a1*y)*Mod(1,2)
W4 = (b2*y^2+b1*y)*Mod(1,2)
W3 = (a2+a1+a0)   *Mod(1,2)
W2 = (b2+b1+b0)   *Mod(1,2)
W1 = W2 * W3                \\ C(1)
W3 = W3 + W0
W2 = W2 + W4
W0 = W0+a0        *Mod(1,2)
W4 = W4+b0        *Mod(1,2)
W3 = W2 * W3                \\ C(y+1)
W2 = W0 * W4                \\ C(y)
W4 = a2 * b2      *Mod(1,2) \\ C(\infty)
W0 = a0 * b0      *Mod(1,2) \\ C(0)

\\ ------ Interpolation phase. 10 add, 2 shift, 2 div. 
W3 = W3 + W2
W3 == ( c4 + (y^2+y+1)*c3 + c2 + c1 )  \\ check

W2 = ( ( W2 + W0 )/y  + W3 + W4*(y^3+1) ) / (y+1)
\\W2 = ( W2 + W0 + W3*y + W4*(y^4+y) ) / (y^2+y)
W2 == ( c2 + c3 )  \\ check

W1 = W1 + W0
W1 == ( c4 + c3 + c2 + c1 )  \\ check

W3 = ( W3 + W1 ) / (y*(y+1))
W3 == ( c3 )  \\ check

W1 = W1 + W4 + W2
W1 == ( c1 )  \\ check

W2 = W2 + W3
W2 == ( c2 )  \\ check

C == W4*x^4+ W3*x^3+ W2*x^2+ W1*x + W0 \\ check

quit;

 */

static
void Toom3Mul (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
              long n, _ntl_ulong *stk)
{
  while (n && a[n-1] == 0 && b[n-1] == 0)
    {
      c[2*n-1] = 0;
      c[2*n-2] = 0;
      n --;
    }

  if (n < MUL_TOOM_THRESHOLD)
    return KarMul (c, a, b, n, stk);

  long k = (n + 2) / 3; /* ceil(n/3) */
  long r = n - 2 * k;
  _ntl_ulong cy;
  _ntl_ulong *W0 = c;
  _ntl_ulong *W1 = stk;
  _ntl_ulong *W2 = c + 2 * k;
  _ntl_ulong *W3 = stk + 2 * k;
  _ntl_ulong *W4 = c + 4 * k;
  _ntl_ulong *newstk = stk + 5 * k +2;

/* \\ --- Evaluation phase. 10 add, 4 shift, 5 mul. */

/* W0 = (a2*y^2+a1*y) */
/* W4 = (b2*y^2+b1*y) */
/*    W0 = (a2*y+a1)*y */
/*    W4 = (b2*y+b1)*y */
  cy = AddLsh1 (W0, a + k, a + 2 * k, r);   /* a1 + x a2 */
  cy = Add1 (W0 + r, a + k + r, k - r, cy);
  W0[k] = (cy << 1) ^ Lsh1 (W0, k, 0); /* x a1 + x^2 a2 */
  cy = AddLsh1 (W4+2, b + k, b + 2 * k, r);
  cy = Add1 (W4+2 + r , b + k + r, k - r, cy);
  W4[k+2] = (cy << 1) ^ Lsh1 (W4+2, k, 0); /* x b1 + x^2 b2 */

/* W3 = ((a2+a1)+a0)    */
/* W2 = ((b2+b1)+b0)    */
  Add (c+k+1,     a, a + k,     k);
  Add (c+k+1, c+k+1, a + 2 * k, r); /* a0 + a1 + a2 */
  Add (W2+2,      b, b + k,     k);
  Add (W2+2,   W2+2, b + 2 * k, r); /* b0 + b1 + b2 */
/* W1 = W2 * W3                \\ C(1) */
  Toom3Mul (W1, c+k+1, W2+2, k, W3); /* W1 : 2*k */

/* W3 = W3 + W0 */
/* W2 = W2 + W4 */
  Add (c+k+1, c+k+1,   W0, k); c[2*k+1]=W0[k];  /* a0 + (x+1)a1 + (x^2+1)a2 */
  Add (W2+2,   W2+2, W4+2, k); W2[k+2]=W4[k+2]; /* b0 + (x+1)b1 + (x^2+1)b2 */

/* W0 = W0+a0    */
/* W4 = W4+b0    */
  Add (W0,     W0, a, k); /* a0 + (x)a1 + (x^2)a2 */
  Add (W4+2, W4+2, b, k); /* b0 + (x)b1 + (x^2)b2 */

/* W3 = W2 * W3                \\ C(y+1) */
/* W2 = W0 * W4                \\ C(y) */
  Toom3Mul (W3, W2+2, c+k+1, k + 1, stk + 5*k + 2); /* W3 : 2*k+1 */
  Toom3Mul (W2,   W0,  W4+2, k + 1, stk + 4*k + 3); /* W2 : 2*k+1 */

/* W4 = a2 * b2                \\ C(\infty) */
/* W0 = a0 * b0                \\ C(0) */
  Toom3Mul (W0, a, b, k, stk + 4*k + 3); /* W0 : 2*k */

  cy = W4[0]; /* Take care of overlapping byte, we just need one bit, but we store all. */
  Toom3Mul (W4, a+2*k, b+2*k,    r, stk + 4*k +3); /* W4 : 2*r */

/* \\ ------ Interpolation phase. 10 add, 2 shift, 2 div.  */

/* W3 = W3 + W2 */
  Add (W3, W3, W2, 2*k);
  W3[2*k] ^= cy;

/* W2 = ( ( W2 + W0 )/y  + W3 + W4*(y^3+1) ) / (y+1) */
/* \\W2 = ( W2 + W0 + W3*y + W4*(y^4+y) ) / (y^2+y) */
  Rsh1Add(W2, W2, W0, 2*k);
  W2[2*k-1] |= cy << (NTL_BITS_PER_LONG - 1);
  /* cy >>= 1; // This byte will be cleared anyway, we can get rid of it. */
/*   Add (W2, W2, W3, 2 * k); */
/*   cy = AddMul1 (W2, W2, W4, 2*r, 9);  */
  Add3 (W2, W3, W4, 2 * r);
  Add (W2 + 2 * r, W2 + 2 * r, W3 + 2 * r, 2 * (k - r));
  cy = AddLsh3 (W2, W2, W4, 2 * r);
  if ( r != k )
    W2[2*r] ^= cy;
  /* else // get rid of the carry, it will be cleared anyway. */
  DivOnePlusX (W2, 2*k);

/* W1 = W1 + W0 */
/* W1 == ( c4 + c3 + c2 + c1 )  \\ check */
  Add (W1, W1, W0, 2 * k);

/* W3 = ( W3 + W1 ) / (y*(y+1)) */
/* W3 == ( c3 )  \\ check */
  Rsh1Add(W3, W3, W1, 2*k);
  W3[2*k-1] |= W3[2*k] << (NTL_BITS_PER_LONG - 1);
  DivOnePlusX (W3, 2*k);

/* W1 = W1 + W4 + W2 */
/* W1 == ( c1 )  \\ check */
  Add3 (W1, W2, W4, 2 * r);
  Add (W1 + 2 * r, W1 + 2 * r, W2 + 2 * r, 2 * (k - r));

/* W2 = W2 + W3 */
/* W2 == ( c2 )  \\ check */
  Add (W2, W2, W3, 2*k);

/* C == W4*x^4+ W3*x^3+ W2*x^2+ W1*x + W0 \\ check */
  Add (c + k, c + k, W1, 4 * k);
}

/* c <- a + x^2 * b, return carry out */
static
_ntl_ulong AddLsh2 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 2) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 2);
      c[i] = t;
    }
  return cy;
}

/* c <- a + x^6 * b, return carry out */
static
_ntl_ulong AddLsh6 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
                    long n)
{
  _ntl_ulong cy = 0UL, t;
  long i;
  for (i = 0; i < n; i++)
    {
      t = a[i] ^ ((b[i] << 6) | cy);
      cy = b[i] >> (NTL_BITS_PER_LONG - 6);
      c[i] = t;
    }
  return cy;
}

/* let c = q*(x+x^4) + X^n*r with X = x^NTL_BITS_PER_LONG and deg(r) < 1
   then c <- q, returns r.
*/
static _ntl_ulong DivExactD1 (_ntl_ulong *c, long n)
{
  _ntl_ulong t = 0;
  long i;

  for (i = 0; i < n; i++) {
    t ^=( c[i] >> 1 ) | ( (i+1<n) ? ( c[i+1] << (NTL_BITS_PER_LONG-1) ) : 0);
    t ^= t << 3 ^ t << 6;
    t ^= t << 9 ^ t << 18;
    t ^= t << 27
#if (NTL_BITS_PER_LONG == 64)
                ^ t << 54
#elif (NTL_BITS_PER_LONG != 32)
#error "NTL_BITS_PER_LONG should be 32 or 64"
#endif
                         ;
    c[i] = t;
    t >>= (NTL_BITS_PER_LONG - 3);
  }
  return t;
}

/* let c = q*(x^2+x^4) + X^n*r with X = x^NTL_BITS_PER_LONG and deg(r) < 1
   then c <- q, returns r.
*/
static _ntl_ulong DivExactD2 (_ntl_ulong *c, long n)
{
  /* c <- c/x, return carry */
/* static */
/* _ntl_ulong Rsh2 (_ntl_ulong *c, long n) */
/* { */
  _ntl_ulong cy = 0, t;
  long i;
  for (i = n - 1; i >= 0; i--)
    {
      t = (c[i] >> 2) | (cy << (NTL_BITS_PER_LONG - 2));
      cy = c[i] & 3UL;
      c[i] = t;
    }
#ifdef DEBUG
  if (cy) abort();
#endif
  return DivOnePlusX2(c,n);
}

#ifdef DEBUG
static dbgcmp(_ntl_ulong *a, const _ntl_ulong *b, long len )
{
  long i;
  for ( i=0; i<len; i++)
    if ( a[i] != b[i] )
      break;
  printf((i==len)?"=":"?");
}

static Dbgcmp(_ntl_ulong *a, const _ntl_ulong *b, long len )
{
  long i;
  for ( i=0; i<len; i++)
    printf((a[i]==b[i])?"+":"-");
}
#endif

/*
\\ gp-pari check code.
\\ (C) 2007 Marco Bodrato <optimaltoom@bodrato.it>
\\ This code is released under GPL 2.0 licence.

U0=u0*Mod(1,2);U1=u1*Mod(1,2);U2=u2*Mod(1,2);U3=u3*Mod(1,2);
V0=v0*Mod(1,2);V1=v1*Mod(1,2);V2=v2*Mod(1,2);V3=v3*Mod(1,2);

U = U3*Y^3 + U2*Y^2*X + U1*Y*X^2 + U0*X^3
V = V3*Y^3 + V2*Y^2*X + V1*Y*X^2 + V0*X^3

\\ P(X,Y): P0=(1,0); P1=(x+1,1); P2=(x,1); P3=(1,1); P4=(1,x); P5=(1,x+1); P6=(0,1)
\\Evaluation phase: 13*2 add, 7*2 shift, 2Smul; 7 mul (n)

W1 = U0 + U1 + U2 + U3    ; W2 = V0 + V1 + V2 + V3
W0 = U1 +(U2 + U3*x)*x    ; W6 = V1 +(V2 + V3*x)*x
W4 = W1 +(W0 + U3*(x+1))*x; W3 = W2 +(W6 + V3*(x+1))*x
W0 = W0*x + U0            ; W6 = W6*x + V0

W5 = W4 * W3              ; W4 = W0 * W6
W3 = W1 * W2

W0 =(U2 +(U1 + U0*x)*x)*x ; W6 =(V2 +(V1 + V0*x)*x)*x
W1 = W1 + W0 + U0*(x^2+x) ; W2 = W2 + W6 + V0*(x^2+x)
W0 = W0 + U3              ; W6 = W6 + V3

W1 = W1 * W2              ; W2 = W0 * W6
W6 = U3 * V3              ; W0 = U0 * V0

\\Interpolation: 22 add, 4 shift, 5 Smul, 4 div (2n)
d1=(x^4+x)*Mod(1,2)	; d1== (x)^1*(x+1)^1*(x^2+x+1)^1 *Mod(1,2)
d2=(x^4+x^2)*Mod(1,2)	; d2== (x)^2*(x+1)^2*(x^2+x+1)^0 *Mod(1,2)

W1 = W1 + W2 + W0*(x^4+x^2+1)
W5 =(W5 + W4 + W1 + W6*(x^4+x^2+1))/d1
W2 = W2 + W6 + W0*(x^6)
W4 = W4 + W2 + W0 + W6*(x^6)
W4 =(W4 + W5*(x^5+x))/d2
W3 = W3 + W0 + W6
W1 = W1 + W3
W2 = W2 +(W1 + W3*x)*x
W3 = W3 + W4 + W5
W1 =(W1 + W3*(x^2+x))/d1
W5 = W5 + W1
W2 =(W2 + W5*(x^2+x))/d2
W4 = W4 + W2

\\Recomposition
W  = W6*Y^6 + W5*Y^5*X + W4*Y^4*X^2+ W3*Y^3*X^3+ W2*Y^2*X^4+ W1*Y*X^5 + W0*X^6
W == U*V
*/
static
void Toom4Mul (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b,
              long n, _ntl_ulong *stk)
{
#ifdef DEBUG
  /* debug */_ntl_ulong *o1, *o2, *rr;
  /* debug */ o1= (unsigned long*) malloc (n * sizeof(unsigned long));
  /* debug */ o2= (unsigned long*) malloc (n * sizeof(unsigned long));
  /* debug */ rr= (unsigned long*) malloc (2*n* sizeof(unsigned long));
#endif

  while (n && a[n-1] == 0 && b[n-1] == 0)
    {
      c[2*n-1] = 0;
      c[2*n-2] = 0;
      n --;
    }

  if (n < MUL_TOOM4_THRESHOLD)
    return Toom3Mul (c, a, b, n, stk);

  long k = (n + 3) / 4; /* ceil(n/4) */
  long r = n - 3 * k;
  _ntl_ulong cy,cy1,cy2,cy4;
  _ntl_ulong *W0 = c;
  _ntl_ulong *W1 = stk;
  _ntl_ulong *W2 = c + 2 * k;
  _ntl_ulong *W3 = stk + 2 * k;
  _ntl_ulong *W4 = c + 4 * k;
  _ntl_ulong *W5 = stk + 4 * k;
  _ntl_ulong *W6 = c + 6 * k;
  _ntl_ulong *newstk = stk + 6 * k + 2;

/* \\Evaluation phase: 13*2 add, 7*2 shift, 2Smul; 7 mul (n) */

/* W1 = U0 + U1 + U2 + U3    ; W2 = V0 + V1 + V2 + V3 */
  Add (W1,     a, a + 3 * k, r);
  Add1(W1+r, a+r, k-r, 0);
  Add3(W1, a + k, a + 2 * k, k); /* U0 + U1 + U2 + U3 */
  Add (W2+2,   b, b + 3 * k, r);
  Add1(W2+r+2,b+r,k-r, 0);
  Add3(W2+2,b+ k, b + 2 * k, k); /* V0 + V1 + V2 + V3 */
/*   Add (W1,     a, a + k,     k); */
/*   Add (W1,    W1, a + 2 * k, k); */
/*   Add (W1,    W1, a + 3 * k, r); /\* U0 + U1 + U2 + U3 *\/ */
/*   Add (W2+2,   b, b + k,     k); */
/*   Add (W2+2,W2+2, b + 2 * k, k); */
/*   Add (W2+2,W2+2, b + 3 * k, r); /\* V0 + V1 + V2 + V3 *\/ */

/* W0 = U1 +(U2 + U3*x)*x    ; W6 = V1 +(V2 + V3*x)*x */
  cy = AddLsh1 (W0, a + 2*k, a + 3 * k, r);   /* U2 + x U3 */
  cy = Add1 (W0 + r, a + 2*k + r, k - r, cy);
  W0[k] = (cy << 1) ^ AddLsh1 (W0, a+k, W0,k); /* U1+x U2 + x^2 U3 */
  cy = AddLsh1 (W6+2, b + 2*k, b + 3 * k, r);   /* V2 + x V3 */
  cy = Add1 (W6+2 + r, b + 2*k + r, k - r, cy);
  W6[k+2] = (cy << 1) ^ AddLsh1 (W6+2, b+k, W6+2,k); /* V1+x V2 + x^2 V3 */
/* W4 = W1 +(W0 + U3*(x+1))*x; W3 = W2 +(W6 + V3*(x+1))*x */
  cy = AddLsh1 (W4+2, W0, a + 3 * k, r);   /* W0 + x U3 */
  cy = Add1 (W4+2+ r, W0 + r, k+1 - r, cy); /* cy == 0 */
  Add (W4+2, W4+2, a + 3 * k, r); /* W0 + x U3 + U3 */
  W4[k+2] = (W4[k+2]<<1)^AddLsh1 (W4+2, W1, W4+2,k); /* W1+x(W0 + (x+1) U3) */
  cy = AddLsh1 (W3+2, W6+2, b + 3 * k, r);   /* W6 + x V3 */
  cy = Add1 (W3+2+ r, W6+2 + r, k+1 - r, cy); /* cy == 0 */
  Add (W3+2, W3+2, b + 3 * k, r); /* W6 + x V3 + V3 */
  W3[k+2] = (W3[k+2]<<1)^AddLsh1 (W3+2,W2+2,W3+2,k); /* W2+x(W6 + (x+1) V3) */
/* W0 = W0*x + U0            ; W6 = W6*x + V0 */
  W0[k] = (W0[k] << 1) ^ AddLsh1 (W0,  a,W0,k); /* U0+x W0 */
  W6[k+2]=(W6[k+2]<<1) ^ AddLsh1 (W6+2,b,W6+2,k); /* V0+x W6 */
/* W5 = W4 * W3              ; W4 = W0 * W6 */
  Toom4Mul (W5, W4+2, W3+2, k+1, newstk); /* W5 : 2*k+1 */
#ifdef DEBUG
  /* debug */ o1[k] =AddMul1(o1,a,a+k,k,3);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+2*k,k,5);
  /* debug */ o1[r]^=AddMul1(o1,o1,a+3*k,r,15);
  /* debug */ o2[k] =AddMul1(o2,b,b+k,k,3);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+2*k,k,5);
  /* debug */ o2[r]^=AddMul1(o2,o2,b+3*k,r,15);
  /* debug */ dbgcmp(o1,W4+2,k+1);
  /* debug */ dbgcmp(o2,W3+2,k+1);
  /* debug */ ToomMul(rr,o1,o2,k+1,rr+n);
  /* debug */ dbgcmp(rr,W5,2*k+1);
#endif
  
  Toom4Mul (W4, W0, W6+2, k+1, newstk); /* W4 : 2*k+1 */
  cy4=W6[0];/* Take care of overlapping byte. */
#ifdef DEBUG
  /* debug */ o1[k] =AddMul1(o1,a,a+k,k,2);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+2*k,k,4);
  /* debug */ o1[r]^=AddMul1(o1,o1,a+3*k,r,8);
  /* debug */ o2[k] =AddMul1(o2,b,b+k,k,2);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+2*k,k,4);
  /* debug */ o2[r]^=AddMul1(o2,o2,b+3*k,r,8);
  /* debug */ dbgcmp(o1,W0,k+1);
  /* debug */ dbgcmp(o2,W6+2,k+1);
  /* debug */ ToomMul(rr,o1,o2,k+1,rr+n);
  /* debug */ dbgcmp(rr,W4,2*k+1);
#endif
/* W3 = W1 * W2 */
  Toom4Mul (W3, W1, W2+2, k, newstk); /* W3 : 2*k */
#ifdef DEBUG
  /* debug */ o1[k] =AddMul1(o1,a,a+k,k,1);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+2*k,k,1);
  /* debug */ o1[r]^=AddMul1(o1,o1,a+3*k,r,1);
  /* debug */ o2[k] =AddMul1(o2,b,b+k,k,1);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+2*k,k,1);
  /* debug */ o2[r]^=AddMul1(o2,o2,b+3*k,r,1);
  /* debug */ dbgcmp(o1,W1,k);
  /* debug */ dbgcmp(o2,W2+2,k);
  /* debug */ ToomMul(rr,o1,o2,k,rr+n);
  /* debug */ dbgcmp(rr,W3,2*k);
#endif

/* W0 =(U2 +(U1 + U0*x)*x)*x ; W6 =(V2 +(V1 + V0*x)*x)*x */
  cy = AddLsh1 (W0, a + 1 * k, a + 0 * k, k);   /* U1 + x U0 */
  W0[k] = (cy << 2) ^ AddLsh1Lsh2(W0, a+2*k, W0,k); /* U2+x U1 + x^2 U0 */
  cy = AddLsh1 (W6+2, b + 1 * k, b + 0 * k, k);   /* V1 + x V0 */
  W6[k+2]=(cy << 2) ^ AddLsh1Lsh2(W6+2, b+2*k, W6+2,k); /* V2+x V1 + x^2 V0 */
/* W1 = W1 + W0 + U0*(x^2+x) ; W2 = W2 + W6 + V0*(x^2+x) */
  W1[k] = AddMul1 (W1, W1, a, k, 4+2);
  Add( W0+k+1, W1, W0, k+1);
  W2[k+2] = AddMul1 (W2+2, W2+2, b, k, 4+2);
  Add( W2+2, W2+2, W6+2, k+1);
/* W0 = W0 + U3              ; W6 = W6 + V3 */
  Add (W0,    W0, a + 3 * k, r); /* + U3 */
  Add (W6+2,W6+2, b + 3 * k, r); /* + V3 */
/* W1 = W1 * W2              ; W2 = W0 * W6 */
  cy=W3[0];cy2=W3[1];/* Take care of overlapping byte. */
  Toom4Mul (W1, W0+k+1, W2+2, k+1, newstk); /* W1 : 2*k+1 */
#ifdef DEBUG
  /* debug */ o1[k] =AddMul1(o1,a ,a+2*k,k,3);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+1*k,k,5);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+0*k,k,0xe);
  /* debug */ o1[r]^=AddMul1(o1,o1,a+3*k,r,1);
  /* debug */ o2[k] =AddMul1(o2,b ,b+2*k,k,3);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+1*k,k,5);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+0*k,k,0xe);
  /* debug */ o2[r]^=AddMul1(o2,o2,b+3*k,r,1);
  /* debug */ dbgcmp(o1,W0+k+1,k+1);
  /* debug */ dbgcmp(o2,W2+2,k+1);
  /* debug */ ToomMul(rr,o1,o2,k+1,rr+n);
  /* debug */ dbgcmp(rr,W1,2*k+1);
#endif
  cy1=W3[0];W3[0]=cy;W3[1]=cy2;
  cy=W4[0];cy2=W4[1];/* Take care of overlapping byte. */
  Toom4Mul (W2, W0, W6+2, k+1, newstk); /* W2 : 2*k+1 */
#ifdef DEBUG
  /* debug */ o1[k] =AddMul1(o1,a ,a+2*k,k,2);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+1*k,k,4);
  /* debug */ o1[k]^=AddMul1(o1,o1,a+0*k,k,9);
  /* debug */ o1[r]^=AddMul1(o1,o1,a+3*k,r,1);
  /* debug */ o2[k] =AddMul1(o2,b ,b+2*k,k,2);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+1*k,k,4);
  /* debug */ o2[k]^=AddMul1(o2,o2,b+0*k,k,9);
  /* debug */ o2[r]^=AddMul1(o2,o2,b+3*k,r,1);
  /* debug */ dbgcmp(o1,W0,k+1);
  /* debug */ dbgcmp(o2,W6+2,k+1);
  /* debug */ ToomMul(rr,o1,o2,k+1,rr+n);
  /* debug */ dbgcmp(rr,W2,2*k+1);
#endif
  W4[1]=cy2;cy2=W4[0];W4[0]=cy;
/* W6 = U3 * V3              ; W0 = U0 * V0 */
  Toom4Mul (W0, a, b, k, newstk); /* W0 : 2*k */
#ifdef DEBUG
  /* debug */ ToomMul(rr, a, b,  k,rr+n);
  /* debug */ dbgcmp(rr,W0,2*k);
#endif
  Toom4Mul (W6, a+3*k, b+3*k,    r, newstk); /* W6 : 2*r */
#ifdef DEBUG
  /* debug */ ToomMul(rr,a+3*k,b+3*k,r,rr+n);
  /* debug */ dbgcmp(rr,W6,2*r);
#endif
/* \\Interpolation: 22 add, 4 shift, 5 Smul, 4 div (2n) */
/* d1=(x^4+x)*Mod(1,2)	; d1== (x)^1*(x+1)^1*(x^2+x+1)^1 *Mod(1,2) */
/* d2=(x^4+x^2)*Mod(1,2)	; d2== (x)^2*(x+1)^2*(x^2+x+1)^0 *Mod(1,2) */

/* W1 = W1 + W2 + W0*(x^4+x^2+1) */

  Add(W1,W1,W2,2*k);
  cy1 ^= cy2 ^ AddMul1(W1,W1,W0,2*k,16+4+1);

/* W5 =(W5 + W4 + W1 + W6*(x^4+x^2+1))/d1 */
  Add3(W5,W4,W1,2*k); W5[2*k] ^= cy1 ^ cy4;
  W5[2*r] ^= AddMul1(W5,W5,W6,2*r,16+4+1);
  DivExactD1(W5,2*k+1);

/* W2 = W2 + W6 + W0*(x^6) */
  Add(W2,W2,W6,2*r);
  cy2^=AddLsh6(W2,W2,W0,2*k);
/* W4 = W4 + W2 + W0 + W6*(x^6) */
  Add3(W4,W2,W0,2*k);cy4^=cy2;
  cy = AddLsh6(W4,W4,W6,2*r);
  cy4^=W6[0];W6[0]^=cy4;cy4^=W6[0]; /* warning: don't overwrite cy [tnx Paul] */
  W4[2*r]^=cy;
/* W4 =(W4 + W5*(x^5+x))/d2 */
  AddMul1(W4,W4,W5,2*k+1,32+2);
  DivExactD2(W4,2*k+1);
  W6[0]=cy4;
/* W3 = W3 + W0 + W6 */
  Add3(W3,W0,W6,2*r);
  if ( r != k )
    Add(W3+2*r,W3+2*r,W0+2*r,2*(k-r)); /* warning: 2r instead of r [tnx Paul] */
/* W1 = W1 + W3 */
  Add(W1,W1,W3,2*k);
/* W2 = W2 +(W1 + W3*x)*x */
  cy2^= AddLsh1(W2,W2,W1,2*k) ^ AddLsh2(W2,W2,W3,2*k) ^ (cy1<<1);
/* W3 = W3 + W4 + W5 */
  Add3(W3,W4,W5,2*k);
/* W1 =(W1 + W3*(x^2+x))/d1 */
  cy=W3[0];
  cy1^=AddMul1(W1,W1,W3,2*k,4+2);
  W3[0]=cy1;
  DivExactD1(W1,2*k+1);
  W3[0]=cy;
/* W5 = W5 + W1 */
  Add(W5,W5,W1,2*k);
/* W2 =(W2 + W5*(x^2+x))/d2 */
  cy=W4[0];
  W4[0]=cy2^AddMul1(W2,W2,W5,2*k,4+2);
  DivExactD2(W2,2*k+1);
  W4[0]=cy;
/* W4 = W4 + W2 */
  Add(W4,W4,W2,2*k);

/* \\Recomposition */
/* W = W6*Y^6 + W5*Y^5 + W4*Y^4+ W3*Y^3+ W2*Y^2+ W1*Y + W0 */
  Add (c + k, c + k, W1, 6 * k);
}
